// Copyright(c) 2017-2018, Intel Corporation
//
// Redistribution  and  use  in source  and  binary  forms,  with  or  without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of  source code  must retain the  above copyright notice,
//   this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name  of Intel Corporation  nor the names of its contributors
//   may be used to  endorse or promote  products derived  from this  software
//   without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,  BUT NOT LIMITED TO,  THE
// IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT  SHALL THE COPYRIGHT OWNER  OR CONTRIBUTORS BE
// LIABLE  FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR
// CONSEQUENTIAL  DAMAGES  (INCLUDING,  BUT  NOT LIMITED  TO,  PROCUREMENT  OF
// SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE,  DATA, OR PROFITS;  OR BUSINESS
// INTERRUPTION)  HOWEVER CAUSED  AND ON ANY THEORY  OF LIABILITY,  WHETHER IN
// CONTRACT,  STRICT LIABILITY,  OR TORT  (INCLUDING NEGLIGENCE  OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#ifndef __FPGA_WSID_LIST_INT_H__
#define __FPGA_WSID_LIST_INT_H__

#include "opae/utils.h"
#include "types_int.h"

/*
 * WSID tracking structure manipulation functions
 */
struct wsid_tracker *wsid_tracker_init(uint32_t n_hash_buckets);
void wsid_tracker_cleanup(struct wsid_tracker *root, void (*clean)(struct wsid_map *));

bool wsid_add(struct wsid_tracker *root,
	      uint64_t wsid,
	      uint64_t addr,
	      uint64_t phys,
	      uint64_t len,
	      uint64_t offset,
	      uint64_t index,
	      int      flags);
bool wsid_del(struct wsid_tracker *root, uint64_t wsid);
uint64_t wsid_gen(void);

struct wsid_map *wsid_find(struct wsid_tracker *root, uint64_t wsid);
struct wsid_map *wsid_find_by_index(struct wsid_tracker *root, uint32_t index);

#endif // ___FPGA_COMMON_INT_H__
